(* Modern reconstruction of the 2006 WG2.11 talk, in the tagless-final terms *)

(* Assume the following signature of a typed AST *)
module type SYM = sig
 type 'a repr
 val int  : int  -> int repr
 val unit : unit -> unit repr
 val app  : ('a->'b) repr -> 'a repr -> 'b repr
 val lam  : ('a repr -> 'b repr) -> ('a->'b) repr
 val (+)  : int repr -> int repr -> int repr
 val (!)  : 'a ref repr -> 'a repr
 val ref  : 'a repr -> 'a ref repr
 val let_ : 'a repr -> ('a repr -> 'b repr) -> 'b repr
 val fseq : unit repr -> 'a repr -> 'a repr     (* sequencing *)
 val ffor : bool -> int repr -> int repr -> 
            (int repr -> unit repr) -> unit repr
 val aref : 'a array repr -> int repr -> 'a repr
 val fass : 'a ref repr -> 'a repr -> unit repr (* assignment *)
 val run  : 'a repr -> 'a code
end

(* Suppose we have a syntactic extension -- ppx, for example, which
re-writes the OCaml code below

funM ENVID n a  ->
  let sum = ref 0 in
  for i = 0 to n do
    sum := !sum + a.(i)
  done;
  !sum

as follows: 
*)


let test1 = fun (module S:SYM) -> let open S in run (
  lam (fun n -> lam (fun a ->
    let_ (ref (int 0)) (fun sum ->
      fseq 
        (ffor true (int 0) n (fun i ->
          fass sum ((! sum) + aref a i)))
        (! sum))))
)
(*
val test1 : (module SYM) -> (int -> int array -> int) code = <fun>
*)

(* Here is one implementation of the SYM, the identity interpretation.
The WG2.11 talk used almost exactly like it.
*)
module ENVID = struct
 type 'a repr = 'a code
 let let_ m f = .<let z = .~m in .~(f .<z>.)>.
 let run x = x
 let int (x:int) = .<x>.
 let unit () = .<()>.
 let app f x = .<.~f .~x>.
 let lam f = .<fun x -> .~(f .<x>.)>.
 let ( + ) x y = .<Pervasives.( + ) .~x .~y>.
 let fseq e1 e2 = .<begin .~e1; .~e2 end>.
 let ffor f e1 e2 body = 
   if f then .<for i = .~e1 to .~e2 do .~(body .<i>.) done>.
        else .<for i = .~e1 downto .~e2 do .~(body .<i>.) done>.
 let ref e = .<ref .~e>.
 let fass e1 e2 = .<(.~e1) := (.~e2)>.
 let ( ! ) e = .<! (.~e) >.
 let aref e1 e2 = .<(.~e1).(.~e2)>.
end

(* In this interpretation, test1 becomes *)
let test1c = test1 (module ENVID)
(*
- : (int -> int array -> int) code = .<
fun x_1  x_2  ->
  let z_3 = Pervasives.ref 0 in
  for i_4 = 0 to x_1 do z_3 := ((! z_3) + (x_2.(i_4))) done; ! z_3>.
*)

open Runcode;;
!. test1c 2 [|1;2;3|];;
(*
- : int = 6
*)

(* Loop unrolling can be done almost exactly as it was in 2006 *)

(* First we define the full loop unrolling combinator *)
let rec full_unroll lb ub body =
  if lb = ub then body lb
  else if lb < ub then .<begin .~(body lb); 
                          .~(full_unroll (lb+1) ub body) end>.
  else .< () >.
;;
(* For example, *)
.<let r = ref 0 in
  .~(full_unroll 1 3 (fun i ->
      .<r := !r + i>.))
>.;;
(* 
- : unit code = .<
let r_9 = Pervasives.ref 0 in
r_9 := ((! r_9) + 1); r_9 := ((! r_9) + 2); r_9 := ((! r_9) + 3)>. 
*)

(* Partial unrolling is expressed in terms of the full unrolling *)

let partial_unroll factor e1 e2 body = 
   .<let lb = .~e1 and ub = .~e2 in
     let ntimes = (ub-lb+1)/factor in
     begin
       for ii = 0 to ntimes-1 do
         .~(full_unroll 0 (factor-1) 
              (fun i -> body .< ii*factor+lb +  i >.))
       done;
       for i = ntimes*factor+lb to ub do .~(body .<i>.) done
     end>.

(* We now re-interpret test1 with partially-unrolled loops *)
let test2c =
  let module ENVUP = 
    struct
      include ENVID
      let ffor = function 
        | true -> partial_unroll 2
        | flag -> ENVID.ffor flag
    end in
  test1(module ENVUP)
;;
(*
val test2c : (int -> int array -> int) code = .<
  fun x_15  x_16  ->
    let z_17 = Pervasives.ref 0 in
    (let lb_18 = 0 and ub_19 = x_15 in
     let ntimes_20 = ((ub_19 - lb_18) + 1) / 2 in
     for ii_22 = 0 to ntimes_20 - 1 do
       (z_17 := ((! z_17) + (x_16.(((ii_22 * 2) + lb_18) + 0)));
        z_17 := ((! z_17) + (x_16.(((ii_22 * 2) + lb_18) + 1))))
     done;
     for i_21 = (ntimes_20 * 2) + lb_18 to ub_19 do
       z_17 := ((! z_17) + (x_16.(i_21)))
     done);
    ! z_17>.
  
*)

!. test2c 2 [|1;2;3|];;
!. test2c 3 [|1;2;3;4|];;
!. test2c 4 [|1;2;3;4;5|];;
